use std::io::{BufRead, Lines};

use super::*;
use itertools::{Itertools, unfold};
use anyhow::{anyhow, Result, Context, bail};

fn range(s: &str) -> Result<(usize, usize, usize)> {
    let (dst, src, len) = s.split_whitespace()
     .collect_tuple()
     .ok_or_else(|| anyhow!("bad range line: {s:?}"))?;
    Ok((dst.parse()?, src.parse()?, len.parse()?))
}

fn seeds(s: &str) -> Result<Vec<usize>> {
    Ok(s.strip_prefix("seeds: ")
        .context("seeds line")?
        .split_whitespace()
        .map(str::parse)
        .collect::<Result<Vec<_>,_>>()?)
}

fn map<R: BufRead>(s: &mut Lines<R>) -> Option<Result<Map>> {
    let title = s.next()?;
    Some((|| {
        let title = title?;
        let (from, to) = title.strip_suffix(" map:")
            .ok_or(anyhow!("bad map header format"))?
            .split_once("-to-")
            .ok_or(anyhow!("bad map name"))?;
        
        let mut ranges = BTreeMap::new();

        loop {
            let line = s.next().ok_or(anyhow!("Unterminated map"))??;
            if line.is_empty() { break }
            let (dst, src, len) = range(&line)?;

            let mut collisions = ranges.range(src..src+len);
            if let Some(old) = collisions.next() {
                if old.0 != old.1 || collisions.next().is_some() {
                    Err(anyhow!("range collision {src}-{len} vs {old:?}"))?;
                }
            }
            ranges.insert(src, dst);
            ranges.entry(src+len).or_insert(src+len);
        }

        Ok(Map { from: from.to_owned(), to: to.to_owned(), ranges })

    })())
}

pub fn input<R: BufRead>(s: R) -> Result<( Vec<usize>, Vec<Map>)> {
    let mut s = s.lines();

    let seeds = seeds(&s.next().ok_or(anyhow!("EOF"))??)?;
    if !s.next().ok_or(anyhow!("EOF"))??.is_empty() {
        bail!("Missing empty line after seeds")
    }

    let maps = unfold(s, map).collect::<Result<_>>()?;

    Ok((seeds, maps))
}
